class AlreadyExistsError(Exception):
    pass


class DataError(Exception):
    pass


class DuplicateError(Exception):
    pass


class MissingAttributeError(Exception):
    pass


class MissingDataError(Exception):
    pass
